*:*****************************************************************************
*:
*: Procedure file: D:\FPD26\EZ\EZ_IBROW.PRG
*:
*:         System: EZ Data Manager
*:         Author: Jack Hairston (209) 572-2634
*:      Copyright (c) 12-25-94, Jack Hairston
*:  Last modified: 12/23/94     19:23
*:
*:  Procs & Fncts: BRWS_CLR
*:               : IBROW1()
*:               : IBROW2
*:               : IBROW3
*:               : IBROW4
*:
*:          Calls: EZ_MSTR.PRG
*:               : WIN.PRG
*:               : CLD.PRG
*:               : BRWS_CLR           (procedure in EZ_IBROW.PRG)
*:               : IBROW1()           (function  in EZ_IBROW.PRG)
*:               : IBROW2             (procedure in EZ_IBROW.PRG)
*:               : IBROW3             (procedure in EZ_IBROW.PRG)
*:               : IBROW4             (procedure in EZ_IBROW.PRG)
*:               : EZ_DBCLK.PRG
*:
*:           Uses: IBROW.DBF          
*:
*:      Documented 12/25/94 at 19:40               FoxDoc  version 2.10f
*:*****************************************************************************
*PROCEDURE ez_ibrow
* Syntax  | search = ibrow([ browse_title [,search_string ;
*         |                [,extra_field [, tag_name [, no_message ]]]]])
* Action  | Incremental browse
*         | As you type in a string, the browse window matches it.
*         | [UP]       Previous browse line
*         | [DOWN]     Next browse line
*         | [HOME]     Clear the search string
*         | [BACK-SP]  Erase latest search character
*         | [ESC]      Returns a null string
*         | [ENTER]    Returns the existing string partially-matched,
*         |            with the file positioned at the record.
*         | Full field Returns the whole string
* Returns | New string "N" + the string
*         | Existing   "Y" + the string (due to partial match)
* Note    | The calling program must handle stripping off this header.
*         |
* Env In  | The current data file must be indexed by .CDX,
*         | or a single .IDX.
*         | xTimeOut   Seconds before time out      default = 300 seconds
*         | xFontMono  Monospaced font              default = "Courier New"
*         | xFontProp  Proportionally-spaced font   default = "Arial"
*         | xFontSize  Size of font                 default = Mac 11, DOS 8
*         | 
* Concept | George Goley, Frank Donovan & Malcolm Rubel
*         | from "FoxPro Power Tools" page 1039
*         |
* Process | lcTitl     Title of the browse window
*         | lcSrch     String to search for    (default to null)
*         | lcExtra    Name of extra field to display  (default null)
*         | lcTag      Name of the tag to search by    (default null = any)
*         | llNoMsg    TRUE to suppress extra message about controls
*         |            FALSE for message  (default)
*         | lcAlias    Name of file to browse  (ALIAS() when called)
*         | lnSrch     Width of search field
*         | lcFd1      Name of key field to browse
*         |  :         (Fields for 2-9 are created if key fields exist.)
*         | lcFd9      Name of field to browse (default to .F.)
*         | lnFlds     Number of fields to browse
*         | lcFld      Name of field
*         | lcOrder    Original index tag
*         | lnTags     Number of index tags for this file
*         | laTag      Array of tags for pick list
*         |      [*,1] Name of the tag
*         |      [*,2] Key() of that tag
*         | lnTag      Pointer to the array of tags
*         | lcPick     Index expression to pick from
*         | lcNear     Current SET NEAR value
*         | lcDel      Current SET DELETED value
*         | lnTop      Top display row
*         | lnBott     Bottom display row
*         | lnOffset   Number of columns to offset for search window
*         | laRray     Array of field names
*         | lnRray     Pointer to the array
*         | lcString   String to search
*         | lnWid      Sum of field widths
*         | lnLast     Last character to check
*         | lnLeft     Left-most display column
*         | lnRight    Right-most display column
*         | lnRows     Number of undeleted records in file, up to 60
*         | lcFontName Name of font to write
*         | lnFontSize Size of font to write
*         | lnRatioHt  Ratio of font heights  (screen:window)
*         | lnRatioWd  Ratio of font widths   (screen:window)
*         | lnMaxCols  Maximum window columns available
*         | lnMaxRows  Maximum window rows available
*         |
* Note  1 | The left-most column of the browse appears empty, but it is
*         | crucially important.  That's where the keyboard input goes,
*         | but is not seen because its foreground color is the same as
*         | the background (DOS).
*         |
*       2 | A clever, bad idea:  Don't add a COLOR clause to any of the
*         | BROWSE statements.  EZ_IBROW depends on the color control in
*         | BRWS_CLR().
*         |
*       3 | If EZ_IBROW() does not wait for input, you probably filled the
*         | initial search field.  For instance, if you want to start
*         | searching with item "HORSE", and the field holds only five
*         | characters, you need to limit the initial search string to
*         | four characters or less, so EZ_IBROW() won't think you have
*         | finished searching.
*         | 
*       4 | Return values:  Since EZ_IBROW() can search on any index tag,
*         | don't depend on the value returned from EZ_IBROW().  Instead,
*         | get the return value from the current record of the parent 
*         | file directly.  For instance:
*         |   lcSrch = ez_ibrow( etc...)
*         |   IF m.lcSrch = "Y"             && successful search
*         |     KeyValue = xyz01.KeyValue   && save the parent value
*         |   ENDIF
*         | 
* Enhance | Allow input of "?" wild card characters, and set filter to
*         | list only data which matches the partial string.
*         | In order to allow spaces, load ibrow.char with an invisible
*         | (in all screen fonts) non-space character.
*         |
*       2 | Add an optional FOR parameter (IsErr = .T.), so that when
*         | browsing for a condition, records would disappear from the
*         | browse as the record was corrected (IsErr becomes .F.).
*         |
*       3 | How to display keys which are wider than the screen?
*         | The ideal would be to display part of each key field,
*         | without exceeding the screen width.
*         |
* History | 08-27-92  jeh  Change procedure to function, returning match.
*         |                Change hard-code to parameter-driven.
*         | 09-23-92  jeh  Moved CURSRCH window from line.
*         |                Moved CURSRCH prompt from left side to right.
*         |                Shortened browse window by one line.
*         |                Added MESSAGE about key controls.
*         |                Changed color scheme to 5 for contrast.
*         | 09-24-92  jeh  Add message about controlling keys.
*         | 09-25-92  jeh  [END] goes to bottom of file
*         | 10-28-92  jeh  Make sure EZ_IBROW.DBF has a record.
*         |                Load IBROW.DBF record with CHR(255)
*         |                Avoid recursive calls
*         | 04-13-94  jeh  Change IBROW.DBF to a cursor.
*         |                Determine the index key on the fly.
*         |                Allow more than one key field.
*         | 04-19-94  jeh  Clear screen message when done.
*         | 04-20-94  jeh  LOCATE before COUNT, to fix bug.
*         | 04-21-94  jeh  Add extra-display-field parameter.
*         | 04-22-94  jeh  Add tag-order pick list.
*         | 04-23-94  jeh  If ESC from order-pick, return [], not .F.
*         | 07-20-94  jeh  Trap no-fields error.
*         |                Include final index field, even if no trailing ")" in KEY().
*         |                Add key parameter, to specify index key for display.
*         | 07-21-94  jeh  Load IBROW.DBF with invisible character--[SPACE].
*         | 07-27-94  jeh  Check TYPE() of all parameters.
*         | 07-28-94  jeh  Drop trailing spaces from initial search argument.
*         | 08-02-94  jeh  Trap invalid tag parameter.
*         | 09-02-94  jeh  Modify code to current standards.
*         | 10-20-94  jeh  Choose a record with double-click (ez_dbclk()).
*         | 12-15-94  jeh  Allow for font differences.
*         | 12-26-94  jeh  Shrink wide data to fit the screen.
*         | 01-03-95  jeh  Default to current screen font.
*         | 02-22-95  jeh  If extra field would be a dupe, ignore it.
*         |                For _MAC, use WROWS() instead of SROWS(),
*         |                  and WCOLS() instead of SCOLS().
*         | 04-02-95  jeh  With [ENTER] remapped to Ctrl-W, pressing [Ctrl-ENTER]
*         |                keyboards a plain Enter, and same for [BACKSPACE].
*--


PARAMETERS lctitl, lcsrch, lcextra, lctag, llnomsg
PRIVATE ALL LIKE l*

*-- Do not clear function keys.  Just define more, then undefine them.
PUSH KEY                               && save function key assignments
CLEAR TYPEAHEAD                        && flush the keyboard buffer


*-- Check parameters and environment ----------------------------------------
IF PARAMETERS() < 4
  lctag = []                           && no key selected
  IF PARAMETERS() < 3
    lcextra = []                       && no extra field
    IF PARAMETERS() < 2
      lcsrch = []                      && null search string
      IF PARAMETERS() < 1
        lctitl = ALIAS()               && file alias
      ENDIF
    ENDIF
  ENDIF
ENDIF

*-- Latest key must not be escape, or Ctrl-W
KEYBOARD "{enter}"                     && reset LASTKEY()...
READ                                   && ...to [ENTER]

*-- All parameters must be type-character
IF TYPE( "m.lcTitl" ) != "C"
  WAIT WINDOW NOWAIT "EZ_IBROW: Window title (1st parameter) must be character"
  = ib_ab_end()                        && abnormal end
ENDIF

IF TYPE( "m.lcSrch" ) != "C"
  WAIT WINDOW NOWAIT "EZ_IBROW: Search argument (2nd parameter) must be character"
  = ib_ab_end()                        && abnormal end
ELSE
  lcsrch = RTRIM( m.lcsrch )           && drop trailing spaces
ENDIF

IF TYPE( "m.lcExtra" ) != "C"
  WAIT WINDOW NOWAIT "EZ_IBROW: Extra field name (3rd parameter) must be character"
  = ib_ab_end()                        && abnormal end
ENDIF

IF TYPE( "m.lcTag" ) != "C"
  WAIT WINDOW NOWAIT "EZ_IBROW: Tag name (4th parameter) must be character"
  = ib_ab_end()                        && abnormal end
ENDIF

*-- There must be a file
IF TYPE( "m.xtimeout" ) = "U"
  xtimeout = 300                       && default to 5 minutes
ENDIF
IF TYPE( "m.xfontprop" ) = "U"
  xfontprop = "Arial"
ENDIF
IF TYPE( "m.xfontmono" ) = "U"
  xfontmono = "Courier New"
ENDIF
IF TYPE( "m.xfontsize" ) = "U"
  xfontsize = IIF( _MAC, 11, 8 )
ENDIF

IF EMPTY( ALIAS())
  = ez_msg( "@*HT \!\<Ok", ;
    "Setup Error", ;
    "No file in use.", ;
    "EZ_IBROW() can't continue." )

  = ib_ab_end()                        && abnormal end
ENDIF

*-- The file must be indexed to match the search field
IF EMPTY( KEY( 1 ))
  = ez_msg( "@*HT \!\<Ok", ;
    "Setup Error ", ;
    "File not indexed: " + ALIAS(), ;
    "EZ_IBROW() can't continue." )

  = ib_ab_end()                        && abnormal end
ENDIF



*-- Initialize memvars --------------------------------------------------
lctopic    = "Incremental Browse"
SET TOPIC TO m.lctopic                 && because of the embedded space
lcalias    = ALIAS()                   && file to browse
lcorder    = ORDER()                   && original index order
lctag      = UPPER( m.lctag )
lntop      = IIF( _DOS, 4, 3.5 )       && top display row


*-- Set fonts for Windows/Macintosh
lnolddec     = SET( "decimal" )
IF _WINDOWS OR _MAC
  SET DECIMAL TO 3
  lnfontsize = m.xfontsize             && default font size
  lcfontname = m.xfontmono             && browse font

  *-- Figure the ratio between screen font and the new font.
  *-- The null string refers to the FoxPro screen.
  lnratioht  = FONTMETRIC( 1, m.lcfontname, m.lnfontsize ) / ;
    FONTMETRIC( 1, WFONT( 1, [] ), WFONT( 2, [] ))
  lnratiowd  = FONTMETRIC( 6, m.lcfontname, m.lnfontsize ) / ;
    FONTMETRIC( 6, WFONT( 1, [] ), WFONT( 2, [] ))
ELSE
  lnratioht  = 1                       && ratio for non-GUI is 1:1
  lnratiowd  = 1
ENDIF

*-- Figure maximum browse window size
DO CASE
CASE _MAC OR _WINDOWS
  *-- To find the size of the FoxPro Mac screen, use WROWS() with a
  *-- null string for the name of the window: WROWS( [] ).
  *-- WROWS() with no window name gives the size of the WONTOP().
  *-- SROWS() gives the size of the physical screen (monitor)
  *-- if SET MACDESKTOP ON, or the size of the FoxPro screen
  *-- if SET MACDESKTOP OFF.
  lnMaxRows = ( WROWS( [] ) / m.lnRatioHt ) - 6
  lnMaxCols = ( WCOLS( [] ) / m.lnRatioWd ) -10
OTHERWISE
  lnmaxrows = SROWS() -6
  lnmaxcols = SCOLS() -10
ENDCASE




*-- Pick an index order, if more than one exists----------------------
*-- unless selected by parameter.
*-- Note:  ORDER() gives the name of the current tag
*--        TAG()   gives the name of the tag for each tag number
*--        KEY()   gives the index expression for each tag number
lntags = 1
DO WHILE ! EMPTY( TAG( m.lntags )) ;
    AND TAG( m.lntags ) != "DEL3TED"   && until no more valid tags
  
  lntags = m.lntags + 1                && how many tags
ENDDO

lntags = m.lntags - 1
DECLARE latag[ m.lnTags, 2 ]

FOR lni = 1 TO m.lntags                && fill array with tags, keys
  latag[ m.lni, 1 ] = UPPER( TAG( m.lni ))
  latag[ m.lni, 2 ] = KEY( m.lni )
ENDFOR

*-- Pick a tag to order displayed data
DO CASE
CASE ! EMPTY( m.lctag )                && a tag was specified
  lnelement = ASCAN( latag, m.lctag )  && find the tag's element
  IF m.lnelement = 0                   && not found
    WAIT WINDOW NOWAIT "EZ_IBROW: Invalid tag (" + m.lctag + ") parameter from " + cld()
    SET DECIMAL TO m.lnolddec          && restore decimal setting
    = ib_ab_end()                      && abnormal end
  ELSE
    lntag = ASUBSCRIPT( latag, m.lnelement, 1 )  && row of index tag
  ENDIF
  
CASE EMPTY( m.lctag ) AND m.lntags > 1 && no tag, and more than one exists
  = win(10, ;
    10, ;
    MIN( ALEN( latag, 1 )+8, m.lnMaxRows -5 ), ;
    IIF(_DOS, 24, 25), ;
    "Pick By")
  
  *-- Pick one of the tags
  @ 0,0 GET   m.lntag ;
    FROM      latag ;
    DEFAULT   1 ;
    PICTURE   "@&T " ;
    SIZE      WROWS(), WCOLS() ;
    MESSAGE   "Choose an index tag to order the pick list." ;
    COLOR     ,,,,,RGB( 255, 255, 255, 128, 0, 0 )

  READ CYCLE MODAL
  RELEASE WINDOW ( WONTOP())
  
  IF LASTKEY() = 27 OR LASTKEY() = 17
    = ib_norm_end()                    && normal end of processing
    RETURN []                          && null string = abandoned
  ELSE
    lcpick  = latag[ m.lnTag, 2 ]      && index expression to pick from
  ENDIF
OTHERWISE
  lntag = 1                            && default to the only tag
ENDCASE

SET ORDER TO m.lntag
lcnear = SET( "near" )                 && save old settings for cleanup
SET NEAR ON
lcdel  = SET( "deleted" )
SET DELETED ON
lcesc = SET( "escape" )
SET ESCAPE ON


*-- Set up temporary work file
CREATE CURSOR ibrow ( char C( 1 ))
INSERT INTO ibrow ( char ) VALUES ( CHR( 0 ))
SELECT ( m.lcalias )                   && data to browse

*-- User message about controls and search string input
lcwind = WONTOP()                      && name of current window
DO CASE
CASE _DOS OR _WINDOWS
  ACTIVATE SCREEN
  IF "DTOS(" $ KEY( m.lntag ) OR "STR(" $ KEY( m.lntag )
    @ SROWS()-2,0 SAY PADC("Enter dates as YYYYMMDD, and " + ;
      "numbers with leading spaces", m.lnMaxCols)
  ENDIF
  @ SROWS()-1,0 SAY PADC("Type the search string   " + ;
    "[ENTER] selects   [ESC] abandons", m.lnMaxCols)  && screen controls
CASE _MAC
  *-- Note:  WROWS( [] ) gives the size of the FoxPro screen.
  *--        The null string is important.
  *--    2:  Documentation would lead you to believe that the correct
  *--        command is ACTIVATE WINDOW SCREEN, but that activates
  *--        the WONTOP() window. <sigh>
  ACTIVATE SCREEN
  IF "DTOS(" $ KEY( m.lntag ) OR "STR(" $ KEY( m.lntag )
    @ WROWS( [] )-2,0 SAY PADC("Enter dates as YYYYMMDD, and " + ;
      "numbers with leading spaces", m.lnMaxCols)
  ENDIF
  @ WROWS( [] )-1,0 SAY PADC("Type the search string   " + ;
    "[ENTER] selects   [ESC] abandons", m.lnMaxCols)  && screen controls
ENDCASE

IF ! EMPTY( m.lcwind )                 && empty if no window
  ACTIVATE WINDOW ( m.lcwind )         && restore current window
ENDIF

***  Change the browse color scheme for hiding entry
*--  The cursor must have fore-/background the same, to hide it.

org_clr   = SCHEME(5)                  && save original scheme
brwse_clr = brws_clr( m.org_clr )      && get and set new one
SET COLOR OF SCHEME 5 TO ( m.brwse_clr )

***  Create browse windows
*-- Note:  You might wonder about the empty-looking field to the
*--        left of the field you are interested in.  The "empty"
*--        field belongs to ibrow->char.  Therefore ib_win is sized
*--        to provide a symmetrical empty field on the right.
*--     2  Only 60 rows are counted because that's bigger than
*--        the maximum screen size--faster than counting all records.

LOCATE                                 && faster than GO TOP
COUNT NEXT 60 TO lnrows                && undeleted records, enough to fill screen

*-- The search window is offset by several columns from the browse window.
DO CASE
CASE _DOS OR _UNIX
  lnoffset  = 3                        && number of columns to offset
CASE _WINDOWS OR _MAC
  lnoffset  = 3 * m.lnratiowd          && proportional spaces are narrower
ENDCASE

*lnbott  = MIN( m.lnMaxRows, m.lntop + m.lnrows + 3 ) ;
*  / m.lnratioht                        && bottom display row
lnbott  = MIN( m.lnMaxRows, m.lntop + m.lnrows + IIF( _DOS, 3, 4 )) && bottom display row


*-- Get the number of fields to browse,
*-- and the names of the fields, from the index KEY()
*-- Note: Field names start with alpha.
*--       Field names contain alpha, number, or underscore only.
*--       Ignore these characters: parens, comma, plus, and space.
= AFIELDS( larray )                    && all field names
lcstring  = KEY( m.lntag )             && string to search
lnlast    = 0                          && last char to check
lnflds    = 0                          && number of fields counter
lnwid     = 0                          && sum of all field widths
lcexclude = ")(, +"                    && excluded characters

DO WHILE m.lnlast < LEN( m.lcstring )
  lnlast = m.lnlast + 1
  
  *-- When the last character is excluded, it's the end of a
  *-- word-to-check.  The word could be a FoxPro function.
  
  IF m.lnlast = LEN( m.lcstring ) OR SUBSTR( m.lcstring, m.lnlast, 1 ) $ m.lcexclude
    IF m.lnlast > 1
      lnrray = ASCAN( larray, LEFT( m.lcstring, m.lnlast -1 ))
    ELSE
      lnrray = 0                       && no-match
    ENDIF
    
    IF m.lnrray > 0                    && found a field name
      lnflds = m.lnflds + 1
      lcfld  = "lcFd" + STR( m.lnflds, 1 )  && name of field variable
      &lcfld = larray[ m.lnRray ]           && field name in column 1
      
      *-- Save the wider of the data, or the field name.
      *-- Field name   is in column 1 of the array.
      *-- Field length is in column 3 of the array.
      lclen  = "lnLn" + STR( m.lnflds, 1 )  && name of length variable
      &lclen = MAX( larray[ m.lnrray +2 ], LEN( larray[ m.lnrray ]))
      
      lnwid = m.lnwid + &lclen
      
    ELSE
      *-- Chop off everything to here
      lcstring = SUBSTR( m.lcstring, m.lnlast +1 )
      lnlast = 0                       && reset to start-of-string
    ENDIF
  ENDIF
ENDDO

*-- If an optional field parameter was passed, include it now,
*-- but only if the field name is valid,
*-- and only if less than nine fields have been defined,
*-- and only if the extra field is not already included.
*-- Note:  Don't combine the next two IF...ENDIFs into a CASE,
*--        because both might be needed.

IF ! EMPTY( m.lcextra )
  FOR lntest = 1 TO m.lnflds
    lctest = "lcFd" + STR( m.lntest, 1 )
    IF m.lcextra = &lctest
      lcextra = []                     && null string--ignore
      EXIT
    ENDIF
  ENDFOR
ENDIF

IF ! EMPTY( m.lcextra ) ;
    AND m.lnflds < 9
  lnflds = m.lnflds + 1                && one more field
  lcfld  = "lcFd" + STR( m.lnflds, 1 )
  &lcfld = m.lcextra                   && name of the extra field
  
  *-- Remember which is wider: Data, or its name?
  lclen  = "lnLn" + STR( m.lnflds, 1 )
  &lclen = MAX( LEN( &lcextra ), LEN( m.lcextra ))
ENDIF

*-- Error:  No fields to browse
IF m.lnflds = 0
  = ez_msg( "@*HT \!\<Ok", ;
    "Setup Error", ;
    "No fields to browse.", ;
    "EZ_IBROW() can't continue." )

  SET DECIMAL TO m.lnolddec            && restore decimal setting
  = ib_ab_end()                        && abnormal end
ENDIF

*-- Calculate the widest window needed to show all columns.
*--   INT()           Make the answer an integer
*--   SCOLS()/2       Center of the screen
*--   EVAL( lcExtra ) Sum the widths of the key fields
*--   2*(lnFlds-1)    Allow for separator grid between columns
lnwid   = m.lnwid + ( 2 * ( m.lnflds -1 )) + ;
  IIF( EMPTY( m.lcextra ), 0, LEN( EVALUATE( m.lcextra )))


*-- Make sure the window is small enough to fit the screen.
*-- Character fields may be made narrower.
*-- All other types are narrow enough to fit (10 bytes or less).
*-- The length of each field is held in lnLn1, lnLn2...
DO WHILE m.lnwid > m.lnmaxcols         && window wider than screen
  lnwid2  = 6                          && allow for outer borders
  FOR lni = 1 TO m.lnFlds              && check all fields
    lcln = "lnLn" + STR( m.lni, 1 )    && name of the width memvar
    IF &lcln > 10                      && character fields only
      &lcln = INT( &lcln * 0.9 )       && shrink it by 10%
    ENDIF
    lnwid2 = lnwid2 + &lcln + 1        && sum the new width + separator
  ENDFOR

  IF m.lnwid2 <= m.lnmaxcols           && these dimensions will fit
    *-- Save the results, without outer borders or final separator
    lnwid = m.lnwid2 -7
  ENDIF
ENDDO


*-- Calculate the right column number
*--   lnLeft          Start at the left-most column
*--   6               Allow for invisible column and borders
*--   LEN(EVAL())     Sum the widths of the fields
*-- Find the left-most window column for display
*-- Note:  Can't MOVE WINDOW...CENTER because the window must
*--        be placed relative to the ib_srch window.
lnleft  = ( WCOLS( [] ) - m.lnwid ) / ( 2 * m.lnRatioWd )
lnright = m.lnleft + 6 + m.lnwid

*-- Note:  Double negatives cause confusion, but parameters default to .F.,
*--        and the message display must be the default condition.
IF ! m.llnomsg
  = win(10,10,22,56,"Incremental Browse Controls" )
  @       1,2  SAY "As you enter characters one at a time," ;
    FONT    m.xfontprop, m.xfontsize
  @ ROW()+1,2  SAY "the browse screen will advance to match." ;
    FONT    m.xfontprop, m.xfontsize
  @ ROW()+1,2  SAY "[Enter] or double-click" ;
    FONT    m.xfontprop, m.xfontsize
  @ ROW()  ,29 SAY "selects" ;
    FONT    m.xfontprop, m.xfontsize
  @ ROW()+1,2  SAY "[Escape]" ;
    FONT    m.xfontprop, m.xfontsize
  @ ROW()  ,29 SAY "cancels" ;
    FONT    m.xfontprop, m.xfontsize
  @ ROW()+1,2  SAY "[Home]" ;
    FONT    m.xfontprop, m.xfontsize
  @ ROW()  ,29 SAY "top of list" ;
    FONT    m.xfontprop, m.xfontsize
  @ ROW()+1,2  SAY "[End]" ;
    FONT    m.xfontprop, m.xfontsize
  @ ROW()  ,29 SAY "bottom" ;
    FONT    m.xfontprop, m.xfontsize
  @ ROW()+1,2  SAY "[Arrows]" ;
    FONT    m.xfontprop, m.xfontsize
  @ ROW()  ,29 SAY "scroll" ;
    FONT    m.xfontprop, m.xfontsize
  WAIT WINDOW NOWAIT "Any key to continue"
  = INKEY( m.xtimeout, "H" )           && wait, hidden cursor
  RELEASE WINDOW (WONTOP())

  IF LASTKEY() = 27                    && escape
    KEYBOARD "{Enter}"
    READ                               && eat the escape
  ENDIF
ENDIF

*-- Define the scroll window of data to match
DEFINE WINDOW ib_win ;
  AT     IIF( _MAC, m.lntop + 0.5, m.lntop ), m.lnleft ;
  SIZE   m.lnbott - m.lntop, m.lnright - m.lnleft ;
  FONT   m.lcfontname, m.lnfontsize ;
  NOCLOSE ;
  NOFLOAT ;
  SHADOW  ;
  COLOR SCHEME 5                       && sized to fit data

SEEK lcsrch                            && position data in the window
DO CASE
CASE lnflds = 9
  *-- Widths must be strings in order to work
  lnLn1 = LTRIM( STR( m.lnLn1 ))
  lnLn2 = LTRIM( STR( m.lnLn2 ))
  lnLn3 = LTRIM( STR( m.lnLn3 ))
  lnLn4 = LTRIM( STR( m.lnLn4 ))
  lnLn5 = LTRIM( STR( m.lnLn5 ))
  lnLn6 = LTRIM( STR( m.lnLn6 ))
  lnLn7 = LTRIM( STR( m.lnLn7 ))
  lnLn8 = LTRIM( STR( m.lnLn8 ))
  lnLn9 = LTRIM( STR( m.lnLn9 ))

  BROWSE FIELDS ibrow->char ;
    :p = "!" ;
    :v = ibrow1( ibrow->char ) ;
    :H = "" ;
    :1, ;
    ( m.lcfd1 ) :&lnLn1, ;
    ( m.lcfd2 ) :&lnLn2, ;
    ( m.lcfd3 ) :&lnLn3, ;
    ( m.lcfd4 ) :&lnLn4, ;
    ( m.lcfd5 ) :&lnLn5, ;
    ( m.lcfd6 ) :&lnLn6, ;
    ( m.lcfd7 ) :&lnLn7, ;
    ( m.lcfd8 ) :&lnLn8, ;
    ( m.lcfd9 ) :&lnLn9  ;
    FONT   m.lcfontname, m.lnfontsize ;
    FREEZE ibrow->char ;
    WINDOW ib_win ;
    TITLE  m.lctitl ;
    NOWAIT NODELETE NOAPPEND NOMENU SAVE
  
CASE lnflds = 8
  *-- Widths must be strings in order to work
  lnLn1 = LTRIM( STR( m.lnLn1 ))
  lnLn2 = LTRIM( STR( m.lnLn2 ))
  lnLn3 = LTRIM( STR( m.lnLn3 ))
  lnLn4 = LTRIM( STR( m.lnLn4 ))
  lnLn5 = LTRIM( STR( m.lnLn5 ))
  lnLn6 = LTRIM( STR( m.lnLn6 ))
  lnLn7 = LTRIM( STR( m.lnLn7 ))
  lnLn8 = LTRIM( STR( m.lnLn8 ))

  BROWSE FIELDS ibrow->char ;
    :p = "!" ;
    :v = ibrow1( ibrow->char ) ;
    :H = "" ;
    :1, ;
    ( m.lcfd1 ) :&lnLn1, ;
    ( m.lcfd2 ) :&lnLn2, ;
    ( m.lcfd3 ) :&lnLn3, ;
    ( m.lcfd4 ) :&lnLn4, ;
    ( m.lcfd5 ) :&lnLn5, ;
    ( m.lcfd6 ) :&lnLn6, ;
    ( m.lcfd7 ) :&lnLn7, ;
    ( m.lcfd8 ) :&lnLn8  ;
    FONT   m.lcfontname, m.lnfontsize ;
    FREEZE ibrow->char ;
    WINDOW ib_win ;
    TITLE  m.lctitl ;
    NOWAIT NODELETE NOAPPEND NOMENU SAVE
  
CASE lnflds = 7
  *-- Widths must be strings in order to work
  lnLn1 = LTRIM( STR( m.lnLn1 ))
  lnLn2 = LTRIM( STR( m.lnLn2 ))
  lnLn3 = LTRIM( STR( m.lnLn3 ))
  lnLn4 = LTRIM( STR( m.lnLn4 ))
  lnLn5 = LTRIM( STR( m.lnLn5 ))
  lnLn6 = LTRIM( STR( m.lnLn6 ))
  lnLn7 = LTRIM( STR( m.lnLn7 ))

  BROWSE FIELDS ibrow->char ;
    :p = "!" ;
    :v = ibrow1( ibrow->char ) ;
    :H = "" ;
    :1, ;
    ( m.lcfd1 ) :&lnLn1, ;
    ( m.lcfd2 ) :&lnLn2, ;
    ( m.lcfd3 ) :&lnLn3, ;
    ( m.lcfd4 ) :&lnLn4, ;
    ( m.lcfd5 ) :&lnLn5, ;
    ( m.lcfd6 ) :&lnLn6, ;
    ( m.lcfd7 ) :&lnLn7  ;
    FONT   m.lcfontname, m.lnfontsize ;
    FREEZE ibrow->char ;
    WINDOW ib_win ;
    TITLE  m.lctitl ;
    NOWAIT NODELETE NOAPPEND NOMENU SAVE
  
CASE lnflds = 6
  *-- Widths must be strings in order to work
  lnLn1 = LTRIM( STR( m.lnLn1 ))
  lnLn2 = LTRIM( STR( m.lnLn2 ))
  lnLn3 = LTRIM( STR( m.lnLn3 ))
  lnLn4 = LTRIM( STR( m.lnLn4 ))
  lnLn5 = LTRIM( STR( m.lnLn5 ))
  lnLn6 = LTRIM( STR( m.lnLn6 ))

  BROWSE FIELDS ibrow->char ;
    :p = "!" ;
    :v = ibrow1( ibrow->char ) ;
    :H = "" ;
    :1, ;
    ( m.lcfd1 ) :&lnLn1, ;
    ( m.lcfd2 ) :&lnLn2, ;
    ( m.lcfd3 ) :&lnLn3, ;
    ( m.lcfd4 ) :&lnLn4, ;
    ( m.lcfd5 ) :&lnLn5, ;
    ( m.lcfd6 ) :&lnLn6, ;
    FONT   m.lcfontname, m.lnfontsize ;
    FREEZE ibrow->char ;
    WINDOW ib_win ;
    TITLE  m.lctitl ;
    NOWAIT NODELETE NOAPPEND NOMENU SAVE
  
CASE lnflds = 5
  *-- Widths must be strings in order to work
  lnLn1 = LTRIM( STR( m.lnLn1 ))
  lnLn2 = LTRIM( STR( m.lnLn2 ))
  lnLn3 = LTRIM( STR( m.lnLn3 ))
  lnLn4 = LTRIM( STR( m.lnLn4 ))
  lnLn5 = LTRIM( STR( m.lnLn5 ))

  BROWSE FIELDS ibrow->char ;
    :p = "!" ;
    :v = ibrow1( ibrow->char ) ;
    :H = "" ;
    :1, ;
    ( m.lcfd1 ) :&lnLn1, ;
    ( m.lcfd2 ) :&lnLn2, ;
    ( m.lcfd3 ) :&lnLn3, ;
    ( m.lcfd4 ) :&lnLn4, ;
    ( m.lcfd5 ) :&lnLn5  ;
    FONT   m.lcfontname, m.lnfontsize ;
    FREEZE ibrow->char ;
    WINDOW ib_win ;
    TITLE  m.lctitl ;
    NOWAIT NODELETE NOAPPEND NOMENU SAVE
  
CASE lnflds = 4
  *-- Widths must be strings in order to work
  lnLn1 = LTRIM( STR( m.lnLn1 ))
  lnLn2 = LTRIM( STR( m.lnLn2 ))
  lnLn3 = LTRIM( STR( m.lnLn3 ))
  lnLn4 = LTRIM( STR( m.lnLn4 ))

  BROWSE FIELDS ibrow->char ;
    :p = "!" ;
    :v = ibrow1( ibrow->char ) ;
    :H = "" ;
    :1, ;
    ( m.lcfd1 ) :&lnLn1, ;
    ( m.lcfd2 ) :&lnLn2, ;
    ( m.lcfd3 ) :&lnLn3, ;
    ( m.lcfd4 ) :&lnLn4  ;
    FONT   m.lcfontname, m.lnfontsize ;
    FREEZE ibrow->char ;
    WINDOW ib_win ;
    TITLE  m.lctitl ;
    NOWAIT NODELETE NOAPPEND NOMENU SAVE
  
CASE lnflds = 3
  *-- Widths must be strings in order to work
  lnLn1 = LTRIM( STR( m.lnLn1 ))
  lnLn2 = LTRIM( STR( m.lnLn2 ))
  lnLn3 = LTRIM( STR( m.lnLn3 ))

  BROWSE FIELDS ibrow->char ;
    :p = "!" ;
    :v = ibrow1(ibrow->char) ;
    :H = "" ;
    :1, ;
    ( m.lcfd1 ) :&lnLn1, ;
    ( m.lcfd2 ) :&lnLn2, ;
    ( m.lcfd3 ) :&lnLn3  ;
    FONT   m.lcfontname, m.lnfontsize ;
    FREEZE ibrow->char ;
    WINDOW ib_win ;
    TITLE  m.lctitl ;
    NOWAIT NODELETE NOAPPEND NOMENU SAVE
  
CASE lnflds = 2
  *-- Widths must be strings in order to work
  lnLn1 = LTRIM( STR( m.lnLn1 ))
  lnLn2 = LTRIM( STR( m.lnLn2 ))

  BROWSE FIELDS ibrow->char ;
    :p = "!" ;
    :v = ibrow1(ibrow->char) ;
    :H = "" ;
    :1, ;
    ( m.lcfd1 ) :&lnLn1, ;
    ( m.lcfd2 ) :&lnLn2  ;
    FONT   m.lcfontname, m.lnfontsize ;
    FREEZE ibrow->char ;
    WINDOW ib_win ;
    TITLE  m.lctitl ;
    NOWAIT NODELETE NOAPPEND NOMENU SAVE
  
CASE lnflds = 1
  *-- Widths must be strings in order to work
  lnLn1 = LTRIM( STR( m.lnLn1 ))

  BROWSE FIELDS ibrow->char ;
    :p = "!" ;
    :v = ibrow1(ibrow->char) ;
    :H = "" ;
    :1, ;
    ( m.lcfd1 ) :&lnLn1 ;
    FONT   m.lcfontname, m.lnfontsize ;
    FREEZE ibrow->char ;
    WINDOW ib_win ;
    TITLE  m.lctitl ;
    NOWAIT NODELETE NOAPPEND NOMENU SAVE
ENDCASE


*--  Create window for display of current entry value
*--  This is our refresh window
*--  Hang the prompt off the left side of the browse window.

lctitle = "Search For"
lnsrch  = LEN( EVALUATE( KEY( m.lntag )))  && width of search field
lntoprow = IIF( _DOS OR _UNIX, 1, 0 )      && higher for Win/Mac
DEFINE WINDOW ib_srch ;
  AT     m.lntoprow, m.lnleft + m.lnoffset ;
  SIZE   1, MAX( m.lnsrch +5, LEN( m.lctitle ) +2 ) ;
  DOUBLE ;
  FONT   m.lcfontname, m.lnfontsize ;
  NOCLOSE ;
  NOFLOAT ;
  TITLE  m.lctitle ;
  COLOR  SCHEME 5

ACTIVATE WINDOW ib_srch
@ 0,0      SAY m.lcsrch + "_" ;
  FONT  m.xfontmono, m.xfontsize       && search string to seek

ON KEY LABEL backspace  DO ibrow2      && remove latest char
ON KEY LABEL ctrl+backspace ;
  KEYBOARD "{backspace}" PLAIN
ON KEY LABEL home       DO ibrow3      && clear the search string
ON KEY LABEL END        DO ibrow4      && go bottom
ON KEY LABEL enter      KEYBOARD CHR(23)  && [^W]
ON KEY LABEL ctrl+enter KEYBOARD "{enter}" PLAIN
ON KEY LABEL MOUSE      DO ez_dbclk    && exit if double-click

*-- loop until ESC or ^W, or until entire key is matched
SET REFRESH TO 10, 5
DO WHILE LASTKEY() != 27 AND LASTKEY() != 17 AND LASTKEY() != 23 ;
    AND LEN( m.lcsrch ) < LEN( EVALUATE( KEY( m.lntag )))
  
  SEEK m.lcsrch                        && position the browse window
  
  IF ! FOUND() AND ! BOF()             && SET NEAR is on
    SKIP - 1                           && nearest match
  ENDIF
  
  BROWSE LAST
ENDDO

= ib_norm_end()                        && normal end for non-failure
RETURN m.lcsrch
*!*****************************************************************************
*!
*!      PROCEDURE: IBROW1
*!
*!      Called by: EZ_IBROW.PRG                      
*!
*!*****************************************************************************
PROCEDURE ibrow1
* Action  | Add latest char to search string.
* Enhance | Figure a way to add custom validation rules,
*         | with custom error messages.
*         |   parameter: valid_function
*         |   flag = IIF(TYPE("valid") = "L", .F., .T.)
*         |   IF flag, to valid()
*--


PARAMETERS thischar
PRIVATE ;
  thischar, ;
  lnkey

*-- Note:  ThisChar is upper case because of the PICTURE clause.
lcsrch = m.lcsrch + m.thischar         && allow spaces
REPLACE ibrow->char WITH CHR( 0 )      && invisible char

DO WHILE .T.                           &&  the timing loop!
  lnkey = INKEY( 0.3, "H" )            &&  wait for second keypress, hidden cursor
  
  DO CASE
  CASE m.lnkey = 0                     && no keypress
    EXIT
  OTHERWISE                            &&  keypress
    *-- Note: These keypresses must be forced upper case,
    *--       because they don't go through the PICTURE clause.
    lcsrch = m.lcsrch + ;
      IIF( BETWEEN( m.lnkey, 32, 122 ),;
      UPPER( CHR( m.lnkey )), [])      && [SPACE] through "z"
  ENDCASE
ENDDO

ACTIVATE WINDOW ib_srch
@ 0,0 SAY m.lcsrch + "_" ;
  FONT  m.xfontmono, m.xfontsize ;
  COLOR RGB( 255, 255, 255, 128, 0, 0 ) && show new search string, w/"cursor"


*!*****************************************************************************
*!
*!      Procedure: IBROW2
*!
*!      Called by: EZ_IBROW.PRG                      
*!
*!*****************************************************************************
PROCEDURE ibrow2
* Action  | [BackSpace] removes latest char from search string.

PUSH KEY CLEAR                         && avoid recursive calls

lcsrch = LEFT( m.lcsrch, LEN( m.lcsrch ) -1 )
ACTIVATE WINDOW ib_srch
IF LEN( m.lcsrch ) < LEN( EVALUATE( KEY( m.lntag ))) - 1
  @ 0,0 SAY m.lcsrch + "_ " ;
    FONT    m.xfontmono, m.xfontsize   && trailing space hides former char
ELSE
  @ 0,0 SAY m.lcsrch + " " ;
    FONT    m.xfontmono, m.xfontsize            && space hides former char
ENDIF

CLEAR TYPEAHEAD                        && drain the keyboard buffer
POP KEY                                && restore all OKL's
*!*****************************************************************************
*!
*!      Procedure: IBROW3
*!
*!      Called by: EZ_IBROW.PRG                      
*!
*!*****************************************************************************
PROCEDURE ibrow3
* Action  | [Home] clears the search string
* Note    | The browse cursor moves to the top of the file.
*--
PUSH KEY CLEAR                         && avoid recursive calls

ACTIVATE WINDOW ib_srch
@ 0,0 SAY "_" + SPACE(LEN( EVALUATE( KEY( m.lntag ))) - 1) ;
  FONT m.xfontmono, m.xfontsize        && hide all former chars
lcsrch = []                            && start over

CLEAR TYPEAHEAD                        && drain the keyboard buffer
POP KEY                                && restore all OKL's
*!*****************************************************************************
*!
*!      Procedure: IBROW4
*!
*!      Called by: EZ_IBROW.PRG                      
*!
*!*****************************************************************************
PROCEDURE ibrow4
* Action  | Move the browse cursor to the bottom of the file,
*         | and clear the search string.
*--

PUSH KEY CLEAR                         && avoid recursive calls
GO BOTTOM
lcsrch = []                            && null search string
CLEAR TYPEAHEAD                        && drain the keyboard buffer
POP KEY                                && restore all OKL's
*!*****************************************************************************
*!
*!      Procedure: BRWS_CLR
*!
*!      Called by: EZ_IBROW.PRG                      
*!
*!*****************************************************************************
PROCEDURE brws_clr
PARAMETERS clr
PRIVATE tst

*** Creates a special color scheme needed by the browse

tst = m.clr

***  Get color pair for OTHER RECORDS             (color pair 1)
***  and make foreground color = background color (invisible)

pos1 = 1
pos2 = AT( ",", tst, 1 )
*new_clr = SUBSTR( tst, pos1 + 1, pos2 -  pos1 - 1)
*new_clr = RIGHT( new_clr, LEN( new_clr ) - AT( "/", new_clr ))
new_clr = "r/r"
new_clr = new_clr + "/" + new_clr

*-- Change the color of the current field (cp 2) to make it invisible.
pos1 = AT( ",", tst, 1 )
pos2 = AT( ",", tst, 2 )
old_pair = SUBSTR( tst, pos1 + 1, pos2 -  pos1 - 1)
tst = STUFF( tst, pos1 + 1, LEN( old_pair ), new_clr )

*-- DOS: Change the current record (color pair 7) to make it stand out.
pos1 = AT( ",", tst, 6 )
pos2 = AT( ",", tst, 7 )
old_pair = SUBST( tst, pos1 + 1, pos2 - pos1 -1 )
tst = STUFF( tst, pos1 + 1, LEN( old_pair ), "rg+/w" )

RETURN(tst)
PROCEDURE ib_norm_end
SET REFRESH TO 0, 300                  && speeds up non-browse processing

DO CASE
CASE LASTKEY() = 27 ;
    OR LASTKEY() = 17 && ESC or Ctrl-Q
  lcsrch = []                          && null string = abandoned
  KEYBOARD "{enter}"                   && clear LASTKEY()...
  = INKEY( 0.2 )                       && ...to 13

CASE LEN( m.lcsrch ) < LEN( EVALUATE( KEY( m.lntag )))  && partial match
  *-- Indicate whether the answer existed before, or not.
  lcsrch = "Y" + EVALUATE( KEY( m.lntag ))  && yes, it existed before
  KEYBOARD "{enter}"                   && clear LASTKEY() = 23  (Ctrl-W)
  = INKEY( 0.2 )                       && ...to 13

OTHERWISE                                   && completely typed by user
  IF SEEK(lcsrch)
    lcsrch = "Y" + m.lcsrch            && yes, it existed before
  ELSE
    lcsrch = "N" + m.lcsrch            && no, it's new
  ENDIF
ENDCASE

RELEASE WINDOWS ib_win, ib_srch        && close local windows

DO CASE
CASE _DOS OR _WINDOWS
  ACTIVATE SCREEN
  @ SROWS()-2, 0 CLEAR                 && remove "message"
CASE _MAC
  *-- Note: Documentation would lead you to believe that the correct
  *--       command is ACTIVATE WINDOW SCREEN, but that activates
  *--       the WONTOP() window. <sigh>
  ACTIVATE SCREEN
  @ WROWS( [] )-2, 0 CLEAR             && remove "message"
ENDCASE

IF USED( "ibrow" )
  USE IN ibrow                         && dump the cursor
ENDIF

SELECT ( m.lcalias )
SET ORDER TO ( m.lcorder )
SET DECIMAL TO m.lnolddec              && restore decimal setting

IF TYPE( "m.lcnear" ) != "U"
  SET NEAR    &lcnear                  && restore SET NEAR
ENDIF

IF TYPE( "m.lcdel" ) != "U"
  SET DELETED &lcdel
ENDIF

IF TYPE( "m.org_clr" ) != "U"
  SET COLOR OF SCHEME 5 TO ( org_clr )   && restore color scheme
ENDIF

IF TYPE( "m .lcesc" ) != "U" AND m.lcesc = "OFF"
  SET ESCAPE OFF                       && restore escape
ENDIF

POP KEY                                && restore former key mapping (OKLs)
PROCEDURE ib_ab_end
  POP KEY                              && restore hot keys (OKLs)
  = ez_mstr()                          && clean up, RETURN TO MASTER
*: EOF: EZ_IBROW.PRG
